from __future__ import absolute_import

from opentracing.ext import tags

from opentracing_instrumentation import utils
from ..request_context import get_current_span
from ._patcher import Patcher


try:
    from boto3.resources.action import ServiceAction
    from botocore import xform_name
    from botocore.exceptions import ClientError
except ImportError:  # pragma: no cover
    pass
else:
    _service_action_call = ServiceAction.__call__


class Boto3Patcher(Patcher):
    applicable = '_service_action_call' in globals()

    def _install_patches(self):
        ServiceAction.__call__ = self._get_call_wrapper()

    def _reset_patches(self):
        ServiceAction.__call__ = _service_action_call

    @staticmethod
    def set_request_id_tag(span, response):
        metadata = response.get('ResponseMetadata')

        # there is no ResponseMetadata for
        # boto3:dynamodb:describe_table
        if metadata:
            span.set_tag('aws.request_id', metadata['RequestId'])

    def _get_call_wrapper(self):
        def call_wrapper(service, parent, *args, **kwargs):
            """Wraps ServiceAction.__call__"""

            service_name = parent.meta.service_name
            operation_name = 'boto3:{}:{}'.format(
                service_name,
                xform_name(service._action_model.request.operation)
            )
            span = utils.start_child_span(operation_name=operation_name,
                                          parent=get_current_span())

            span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
            span.set_tag(tags.COMPONENT, 'boto3')
            span.set_tag('boto3.service_name', service_name)

            with span:
                try:
                    response = _service_action_call(service, parent,
                                                    *args, **kwargs)
                except ClientError as error:
                    self.set_request_id_tag(span, error.response)
                    raise
                else:
                    if isinstance(response, dict):
                        self.set_request_id_tag(span, response)

            return response

        return call_wrapper


patcher = Boto3Patcher()


def set_patcher(custom_patcher):
    global patcher
    patcher = custom_patcher


def install_patches():
    patcher.install_patches()


def reset_patches():
    patcher.reset_patches()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from opentracing.ext import tags as ext_tags
import re

from ._current_span import current_span_func
from ._singleton import singleton
from .. import utils

try:
    import redis
except ImportError:  # pragma: no cover
    redis = None


# regex to match an ipv4 address
IPV4_RE = re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$')

METHOD_NAMES = ['execute_command', 'get', 'set', 'setex', 'setnx']
ORIG_METHODS = {}


@singleton
def install_patches():
    if redis is None:
        return

    def peer_tags(self):
        """Fetch the peer host/port tags for opentracing.

        We do this lazily and cache the result since the host/port won't
        change.
        """
        if not hasattr(self, '_peer_tags'):
            self._peer_tags = []
            conn_info = self.connection_pool.connection_kwargs
            host = conn_info.get('host')
            if host:
                if IPV4_RE.match(host):
                    self._peer_tags.append((ext_tags.PEER_HOST_IPV4, host))
                else:
                    self._peer_tags.append((ext_tags.PEER_HOSTNAME, host))
            port = conn_info.get('port')
            if port:
                self._peer_tags.append((ext_tags.PEER_PORT, port))
        return self._peer_tags

    redis.StrictRedis.peer_tags = peer_tags

    for name in METHOD_NAMES:
        ORIG_METHODS[name] = getattr(redis.StrictRedis, name)

    def get(self, name, **kwargs):
        self._extra_tags = [('redis.key', name)]
        return ORIG_METHODS['get'](self, name, **kwargs)

    def set(self, name, value, ex=None, px=None, nx=False, xx=False, **kwargs):
        self._extra_tags = [('redis.key', name)]
        return ORIG_METHODS['set'](self, name, value,
                                   ex=ex, px=px, nx=nx, xx=xx,
                                   **kwargs)

    def setex(self, name, time, value, **kwargs):
        self._extra_tags = [('redis.key', name),
                            ('redis.ttl', time)]
        return ORIG_METHODS['setex'](self, name, time, value, **kwargs)

    def setnx(self, name, value, **kwargs):
        self._extra_tags = [('redis.key', name)]
        return ORIG_METHODS['setnx'](self, name, value, **kwargs)

    def execute_command(self, cmd, *args, **kwargs):
        operation_name = 'redis:%s' % (cmd,)
        span = utils.start_child_span(
            operation_name=operation_name, parent=current_span_func())
        span.set_tag(ext_tags.SPAN_KIND, ext_tags.SPAN_KIND_RPC_CLIENT)
        span.set_tag(ext_tags.PEER_SERVICE, 'redis')

        # set the peer information (remote host/port)
        for tag_key, tag_val in self.peer_tags():
            span.set_tag(tag_key, tag_val)

        # for certain commands we'll add extra attributes such as the redis key
        for tag_key, tag_val in getattr(self, '_extra_tags', []):
            span.set_tag(tag_key, tag_val)
        self._extra_tags = []

        with span:
            return ORIG_METHODS['execute_command'](self, cmd, *args, **kwargs)

    for name in METHOD_NAMES:
        setattr(redis.StrictRedis, name, locals()[name])


def reset_patches():
    for name in METHOD_NAMES:
        setattr(redis.StrictRedis, name, ORIG_METHODS[name])
    ORIG_METHODS.clear()
    install_patches.reset()



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from future import standard_library

standard_library.install_aliases()
import logging
import urllib.parse

from opentracing.ext import tags
from ..http_client import AbstractRequestWrapper
from ..http_client import before_http_request
from ..http_client import split_host_and_port
from ._patcher import Patcher
from ._current_span import current_span_func

log = logging.getLogger(__name__)

# Try to save the original entry points
try:
    import requests.adapters
except ImportError:  # pragma: no cover
    pass
else:
    _HTTPAdapter_send = requests.adapters.HTTPAdapter.send


class RequestsPatcher(Patcher):
    applicable = '_HTTPAdapter_send' in globals()
    response_handler_hook = None

    def set_response_handler_hook(self, response_handler_hook):
        """
        Set a hook that will be called when a response is received.

        The hook can be set in purpose to set custom tags to spans
        depending on content or some metadata of responses.

        :param response_handler_hook: hook method
            It must have a signature `(response, span)`,
            where `response` and `span` are positional arguments,
            so you can use different names for them if needed.
        """

        self.response_handler_hook = response_handler_hook

    def _install_patches(self):
        requests.adapters.HTTPAdapter.send = self._get_send_wrapper()

    def _reset_patches(self):
        requests.adapters.HTTPAdapter.send = _HTTPAdapter_send

    def _get_send_wrapper(self):
        def send_wrapper(http_adapter, request, **kwargs):
            """Wraps HTTPAdapter.send"""
            request_wrapper = self.RequestWrapper(request=request)
            span = before_http_request(request=request_wrapper,
                                       current_span_extractor=current_span_func
                                       )
            with span:
                response = _HTTPAdapter_send(http_adapter, request, **kwargs)
                if getattr(response, 'status_code', None) is not None:
                    span.set_tag(tags.HTTP_STATUS_CODE, response.status_code)
                if self.response_handler_hook is not None:
                    self.response_handler_hook(response, span)
            return response

        return send_wrapper

    class RequestWrapper(AbstractRequestWrapper):
        def __init__(self, request):
            self.request = request
            self.scheme, rest = urllib.parse.splittype(request.url)
            if self.scheme and rest:
                self.host_str, _ = urllib.parse.splithost(rest)
            else:
                self.host_str = ''

        def add_header(self, key, value):
            self.request.headers[key] = value

        @property
        def method(self):
            return self.request.method

        @property
        def full_url(self):
            return self.request.url

        @property
        def _headers(self):
            return self.request.headers

        @property
        def host_port(self):
            return split_host_and_port(host_string=self.host_str,
                                       scheme=self.scheme)


patcher = RequestsPatcher()


def set_patcher(custom_patcher):
    global patcher
    patcher = custom_patcher


def install_patches():
    patcher.install_patches()
